-------------------------------------------------------------------------------
--  SCREAM core testbench                                                    --
-------------------------------------------------------------------------------
--  Author: L. Gaspar, TTTech Computertechnik AG, Vienna, Austria 2015(C)    --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This testbench verifies the functionality of the SCREAM     --
--               core.                                                       --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 13/12/2015 Lubos Gaspar: decryption support added                 --
--       - 22/07/2014 Lubos Gaspar: first version of SCREAM core TB          --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.declaration_pkg.all;
use work.txt_util.all;

-------------------------------------------------------------------------------
------------------------------- E N T I T Y -----------------------------------
-------------------------------------------------------------------------------
entity tb_Scream128 is
end tb_Scream128;
 
-------------------------------------------------------------------------------
------------------------- A R C H I T E C T U R E -----------------------------
-------------------------------------------------------------------------------
architecture behavior of tb_Scream128 is 

  -- Parameter constants
  constant STEPS  : natural := 10; -- Scream: 10 or 12 steps
  constant UNROLL : boolean := false; -- 1 (if false) or 2 (if true) rounds are impl.    
  
  -- Global signals
  signal clk      : std_logic := '1';
   
  -- Clock period definitions
  constant clk_period   : time := 10 ns;  
  
  -- Input key
  constant KTX : STATE_TYPE := (
  X"1100", -- least significant 16-bit word
  X"3322", 
  X"5544",
  X"7766", 
  X"9988",
  X"BBAA",
  X"DDCC",
  X"FFEE"); -- most significant 16-bit word  
  
  -- Input plaintext
  constant PTX : STATE_TYPE := (
  X"E1F0", -- least significant 16-bit word
  X"C3D2",
  X"A5B4",
  X"8796",
  X"6978",
  X"4B5A",  
  X"2D3C",
  X"0F1E"); -- most significant 16-bit word

  -- Input tweak
  constant TWX : STATE_TYPE := (  
  X"9180", -- least significant 16-bit word
  X"B3A2",
  X"D5C4",
  X"F7E6",
  X"1908",
  X"3B2A",  
  X"5D4C",
  X"7F6E"); -- most significant 16-bit word  
  
  -- Expected ciphertext result for 10-step algorithm
  constant CTX10 : STATE_TYPE := (
  X"78F7",  -- least significant 16-bit word
  X"6EC8", 
  X"A06E",
  X"BC5D",
  X"08D9",
  X"6574",
  X"5AE7",
  X"867F");  -- most significant 16-bit word  
  
  -- Expected ciphertext result for 12-step algorithm
  constant CTX12 : STATE_TYPE := (
  X"A4C6",
  X"41DE",
  X"344E",
  X"F795",
  X"160C",
  X"F72F",
  X"C5C1",
  X"BC8B");
  
  type TYPE_CIP_IN is record
    reset   : std_logic;
    decrypt : std_logic;
    valid   : std_logic;
    mode    : std_logic;
    data    : std_logic_vector(127 downto 0);
    tweak   : std_logic_vector(127 downto 0);
    ready   : std_logic;
  end record;
  
  constant RESET_TYPE_CIP_IN : TYPE_CIP_IN  := (
    reset    => '1',
    decrypt  => '0',
    valid    => '0',
    mode     => '0',
    data     => (others => '0'),
    tweak    => (others => '0'),
    ready    => '1'
    );
    
  type TYPE_CIP_OUT is record
    data    : std_logic_vector(127 downto 0);
    ready   : std_logic;
    valid   : std_logic;    
  end record;   
     
  type TYPE_TB_STATE is record
    plaintext  : std_logic_vector(127 downto 0);
    tweak      : std_logic_vector(127 downto 0);
    key        : std_logic_vector(127 downto 0);
    ciphertext : std_logic_vector(127 downto 0);
  end record;   
  
  signal cipi : TYPE_CIP_IN;
  signal cipo : TYPE_CIP_OUT;
-------------------------------------------------------------------------------
---------------------- T E S T B E N C H   B O D Y ----------------------------
-------------------------------------------------------------------------------
begin
 
-----------------------
--- UNIT UNDER TEST ---
----------------------- 
uut: Scream128encdec
  generic map(
    G_STEPS   => STEPS,
    G_UNROLL  => UNROLL      
  )
  port map(
    clk_i     => clk,    
    rst_i     => cipi.reset,
    -- Choose between encryption or decryption
    decrypt_i => cipi.decrypt,
    -- Input interface side    
    valid_i   => cipi.valid,
    ready_o   => cipo.ready,
    mode_i    => cipi.mode,
    data_i    => cipi.data,
    tweak_i   => cipi.tweak,    
    -- Output interface side
    valid_o   => cipo.valid,
    ready_i   => cipi.ready,
    data_o    => cipo.data
    );
 
-----------------------
--- CLOCK GENERATOR ---
-----------------------   
clk_process:
process
begin
  -- 0/2 of clock cycle
  clk   <= '1';      
  wait for clk_period/2;  
  -- 1/2 of clock cycle  
  clk   <= '0';    
  wait for clk_period/2;   
end process;

------------------------
--- STIMULUS PROCESS ---
------------------------      
stim_proc: 
process
begin
   -- Hold reset state for 20 ns.
   cipi         <= RESET_TYPE_CIP_IN;   

   -- Release reset                 
   wait for 20 ns;
   cipi.reset    <= '0'; 
   wait for 20 ns;      
   
   ------------------------
   --- Encryptions test ---
   ------------------------
   
   -- Write key into the cipher
   wait until (clk = '1');
    cipi.data    <= mx2vec(KTX); -- key data    
    cipi.mode    <= '0'        ; -- set mode to KEY accept
    cipi.valid   <= '1'        ; -- make the key valid for the cipher
      
   -- Write data into the cipher
   wait until (clk = '1');
    cipi.decrypt <= '0'        ; -- encryption
    cipi.data    <= mx2vec(PTX); -- plaintext data
    cipi.tweak   <= mx2vec(TWX); -- tweak 
    cipi.mode    <= '1'        ; -- set mode to PLAINTEXT accept
    cipi.valid   <= '1'        ; -- make the data valid for the cipher
    -- let the cipher wait until the encryption is done 
    -- (TB seems not to be ready to read out the ciphertext)
    cipi.ready   <= '0'        ; 
    
   -- Wait sufficiently long so that the cipher finishes
   wait until (clk = '1');
   while cipo.valid = '0' loop
     wait until (clk = '1');
   end loop;
   
   -- Read out the data
   cipi.ready    <= '1'; 
   
   -- Verify the result   
   if STEPS = 10 then
     if (cipo.data = mx2vec(CTX10)) then
       report "TEST 1 PASSED - RESULT IS CORRECT!";
     else
       report "TEST 1 FAILED - RESULT IS NOT CORRECT!";
     end if;
   else
     if (cipo.data = mx2vec(CTX12)) then
       report "TEST 1 PASSED - RESULT IS CORRECT!";
     else
       report "TEST 1 FAILED - RESULT IS NOT CORRECT!";
     end if;
   end if;
   
   -- Since both cipi.valid and cipi.ready are in 1, 
   -- a new encryption will always start immediately when the 
   -- old is finished.
   
   -- Lets keep encryption running for 1 us
   wait for 0.7 us;
   wait until (clk = '1');
   cipi.valid   <= '0'; -- new encryption will not start when the pending is finished
   wait for 0.3 us;
   
   ------------------------
   --- Decryptions test ---
   ------------------------   
   
   -- Write key into the cipher
   wait until (clk = '1');
    cipi.data    <= mx2vec(KTX); -- key data    
    cipi.mode    <= '0'        ; -- set mode to KEY accept
    cipi.valid   <= '1'        ; -- make the key valid for the cipher
      
   -- Write data into the cipher
   wait until (clk = '1');
    cipi.decrypt <= '1'        ; -- decryption
    if STEPS = 10 then
      cipi.data  <= mx2vec(CTX10); -- ciphertext data
    else
      cipi.data  <= mx2vec(CTX12); -- ciphertext data
    end if;
    cipi.tweak   <= mx2vec(TWX); -- tweak 
    cipi.mode    <= '1'        ; -- set mode to CIPHERTEXT accept
    cipi.valid   <= '1'        ; -- make the data valid for the cipher
    -- let the cipher wait a while when the decryption is done 
    -- (TB seems not to be ready to read out the ciphertext)
    cipi.ready   <= '0'        ; 
    
   -- Wait sufficiently long so that the cipher finishes
   wait until (clk = '1');
   while cipo.valid = '0' loop
     wait until (clk = '1');
   end loop;
   
   -- Read out the data
   cipi.ready    <= '1'; 
   
   -- Verify the result
   if (cipo.data = mx2vec(PTX)) then
     report "TEST 2 PASSED - RESULT IS CORRECT!";
   else
     report "TEST 2 FAILED - RESULT IS NOT CORRECT!";
   end if;
   
   -- Since both cipi.valid and cipi.ready are in 1, 
   -- a new encryption will always start immediately when the 
   -- old is finished.
   
   -- Lets keep encryption running for 1 us
   wait for 1 us;
                                
   ---- END OF TEST ----
      
   wait;
end process;

end;
