-------------------------------------------------------------------------------
--  SCREAM testbench                                                         --
------------------------------------------------------------------------------- 
--  Author: L. Gaspar, TTTech Computertechnik AG, Vienna, Austria 2015(C)    --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This testbench verifies the functionality of the SCREAM     --
--               TAE core.                                                   --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 13/12/2015 Lubos Gaspar: first version of SCREAM TAE              --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.declaration_pkg.all;
use work.txt_util.all;

-------------------------------------------------------------------------------
------------------------------- E N T I T Y -----------------------------------
-------------------------------------------------------------------------------
entity tb_Scream128_tae is
	generic (		
		TRUNC		: boolean := True;
		PADD		: boolean := True;
		STEPS		: natural := 12;
		UNROLL	: boolean := True
	);
end tb_Scream128_tae;
 
-------------------------------------------------------------------------------
------------------------- A R C H I T E C T U R E -----------------------------
-------------------------------------------------------------------------------
architecture behavior of tb_Scream128_tae is  
  
  -- Clock period definitions
  constant clk_period     : time := 10 ns;   
  constant ADATA_BLOCKS   : natural := 3;
  constant ADATA_LAST_LEN : natural := 3;
  constant DATA_BLOCKS    : natural := 6;
  constant DATA_LAST_LEN  : natural := 6;
  
 
  -- Global signals
  signal clk      : std_logic := '1';
    
  type TYPE_IN_DATA128_ARRAY	is array (natural range <>) of std_logic_vector(127 downto 0);
  type TYPE_IN_DATA8_ARRAY    is array (natural range <>) of natural range 0 to 16;
  
  -- 
  signal sim_nounce    : std_logic_vector(127 downto 0); -- only upper 88 bits are used
  signal sim_key       : std_logic_vector(127 downto 0);
  -- 
  signal sim_adata     : TYPE_IN_DATA128_ARRAY(0 to ADATA_BLOCKS-1);
  signal sim_ptext     : TYPE_IN_DATA128_ARRAY(0 to DATA_BLOCKS-1);
  signal sim_ctext     : TYPE_IN_DATA128_ARRAY(0 to DATA_BLOCKS-1);
  signal sim_adata_len : TYPE_IN_DATA8_ARRAY(0 to ADATA_BLOCKS-1);
  signal sim_data_len  : TYPE_IN_DATA8_ARRAY(0 to DATA_BLOCKS-1);
  signal sim_tag       : std_logic_vector(127 downto 0);
  
  constant MODE_NOUNCE : std_logic_vector(1 downto 0) := "00";
  constant MODE_KEY    : std_logic_vector(1 downto 0) := "01";
  constant MODE_ADATA  : std_logic_vector(1 downto 0) := "10";
  constant MODE_DATA   : std_logic_vector(1 downto 0) := "11";
  
  
  type TYPE_CIP_IN is record
    reset   : std_logic;
    decrypt : std_logic;
    valid   : std_logic;
    mode    : std_logic_vector(1 downto 0);
    last    : std_logic;
    len     : natural range 0 to 2**4;
    data    : std_logic_vector(127 downto 0);    
    ready   : std_logic;
  end record;
  
  constant RESET_TYPE_CIP_IN : TYPE_CIP_IN  := (
    reset    => '1',
    decrypt  => '0',
    valid    => '0',
    mode     => "00",
    last     => '0',
    len      => 16,
    data     => (others => '0'),    
    ready    => '1'
    );
    
  type TYPE_CIP_OUT is record
    data    : std_logic_vector(127 downto 0);
    ready   : std_logic;
    valid   : std_logic;
    mode    : std_logic;
    last    : std_logic;
    len     : natural range 0 to 2**4;    
  end record;            
  
  signal cipi : TYPE_CIP_IN;
  signal cipo : TYPE_CIP_OUT;
  signal slv_ilen : std_logic_vector(4 downto 0);
  signal slv_olen : std_logic_vector(4 downto 0);
  
  -- Wait for cipher to accept input data
  procedure wait4cipi is
  begin
    wait until(clk = '1');
    -- loop until cipher is ready
    while (cipo.ready = '0') loop
      wait until(clk = '1');
    end loop;    
  end wait4cipi;
  
  -- Wait for cipher data output to be available
  procedure wait4cipo is
  begin
    wait until(clk = '1');
    -- loop until cipher contains valid data on its output
    while (cipo.valid = '0') loop
      wait until(clk = '1');
    end loop;    
  end wait4cipo;
  

-------------------------------------------------------------------------------
---------------------- T E S T B E N C H   B O D Y ----------------------------
-------------------------------------------------------------------------------
begin
 

process(cipi.reset)
begin
   
  -- Nounce
  sim_nounce   <=  X"0004080C1014181C202428" & X"0000000000"; -- only upper 88 bits are used
  
  -- Key
  sim_key      <=  X"00020406080A0C0E10121416181A1C1E";
  
  -- Associated data
  sim_adata(0) <=  X"FFFEFDFCFBFAF9F8F7F6F5F4F3F2F1F0"; sim_adata_len(0) <= 16;
  sim_adata(1) <=  X"EFEEEDECEBEAE9E8E7E6E5E4E3E2E1E0"; sim_adata_len(1) <= 16;
  sim_adata(2) <=  X"DFDEDDDCDBDAD9D8D700000000000000"; sim_adata_len(2) <=  9;-- truncated intentionally
  
  -- Plaintext vectors
  sim_ptext(0) <=  X"000102030405060708090A0B0C0D0E0F"; sim_data_len(0) <= 16;
  sim_ptext(1) <=  X"101112131415161718191A1B1C1D1E1F"; sim_data_len(1) <= 16;
  sim_ptext(2) <=  X"202122232425262728292A2B2C2D2E2F"; sim_data_len(2) <= 16;
  sim_ptext(3) <=  X"303132333435363738393A3B3C3D3E3F"; sim_data_len(3) <= 16;
  sim_ptext(4) <=  X"404142434445464748494A4B4C4D4E4F"; sim_data_len(4) <= 16;
  sim_ptext(5) <=  X"505152535455565758595A0000000000"; sim_data_len(5) <= 11; -- last block is truncated intentionally (we want to test how cipher behaves)
 
  
  if STEPS = 10 then    
    -- Corresponding ciphertext
    sim_ctext(0) <=  X"77547D776181590AA19C17F22462D3BB";
    sim_ctext(1) <=  X"A5B67E1E7673DAC457CBC7E68F27FBAE";
    sim_ctext(2) <=  X"621CBE2A94959930FFC30A07E28BA914";
    sim_ctext(3) <=  X"F09E16E1EF19636488828DD200CFCE97";
    sim_ctext(4) <=  X"A4D07BFEBC7508E6A64D489F942BE50A";
    sim_ctext(5) <=  X"1851A0C72350ACF26B7B070000000000";
    
    -- Resulting tag
    sim_tag      <=  X"50C3A77EF4E5B5DEE06572F908C927AE";
    
  elsif STEPS = 12 then
    -- Corresponding ciphertext
    sim_ctext(0) <=  X"6FEABBF27E5F2AB013CBE05C8985DFE4";
    sim_ctext(1) <=  X"9F29B8AB3475141996A75E1708D97797";
    sim_ctext(2) <=  X"A4286908EE3D2A019A0CA14F7B5745A2";
    sim_ctext(3) <=  X"077946D675844BC1199149FE2BFC1E19";
    sim_ctext(4) <=  X"FB92EBB39F1AF1600E6E143E0154A120";
    sim_ctext(5) <=  X"BD16987B60EFF0F70EE9D00000000000";
    
    -- Resulting tag
    sim_tag      <=  X"2EC5D54EC1B1E367B6537970399EF7E3";    
  else
    -- TO BE DEFINED FOR OTHER NUMBER OF STEPS
    -- Corresponding ciphertext
    sim_ctext(0) <=  X"00000000000000000000000000000000";
    sim_ctext(1) <=  X"00000000000000000000000000000000";
    sim_ctext(2) <=  X"00000000000000000000000000000000";
    sim_ctext(3) <=  X"00000000000000000000000000000000";
    sim_ctext(4) <=  X"00000000000000000000000000000000";
    sim_ctext(5) <=  X"00000000000000000000000000000000";
    
    -- Resulting tag
    sim_tag      <=  X"00000000000000000000000000000000";   
  end if;

end process;
 
-----------------------
--- UNIT UNDER TEST ---
----------------------- 
scream_tae_top_i: Scream_tae
   generic map(
     TRUNC     => TRUNC,
     PADD      => PADD,
     STEPS     => STEPS,
     UNROLL    => UNROLL
   )
   port map(
     clk       => clk,
     rst       => cipi.reset,

     -- Plaintext interface side
     in_decrypt=> cipi.decrypt,
     in_valid  => cipi.valid,
     in_ready  => cipo.ready, 
     in_mode   => cipi.mode,  
     in_last   => cipi.last,  
     in_data   => cipi.data,  
     in_length => slv_ilen,

     -- Ciphertext interface side
     out_valid => cipo.valid, 
     out_ready => cipi.ready,
     out_mode  => cipo.mode,
     out_last  => cipo.last,  
     out_length=> slv_olen,
     out_data  => cipo.data
     );
 
 slv_ilen  <= std_logic_vector(to_unsigned(cipi.len,5));
 cipo.len  <= to_integer(unsigned(slv_olen));
 
-----------------------
--- CLOCK GENERATOR ---
-----------------------   
clk_process:
process
begin
  -- 0/2 of clock cycle
  clk   <= '1';      
  wait for clk_period/2;  
  -- 1/2 of clock cycle  
  clk   <= '0';    
  wait for clk_period/2;   
end process;

------------------------
--- STIMULUS PROCESS ---
------------------------      
stim_proc: 
process
begin
   -- Hold reset state for 20 ns.
   cipi <= RESET_TYPE_CIP_IN;   

   -- Release reset                 
   wait for 20 ns;
   cipi.reset    <= '0'; 
   wait for 20 ns;      
  
   ------------------------
   --- Encryptions test ---
   ------------------------
   wait until (clk = '1');
   cipi.decrypt  <= '0'                ; -- set encryption
   cipi.ready    <= '1'                ; -- TB can receive all the TAE output data immediately
   
   -- Write nonce into the cipher   
    cipi.mode    <= MODE_NOUNCE        ; -- set mode to NOUNCE
    cipi.valid   <= '1'                ; -- make the nounce valid for TAE
    cipi.data    <= sim_nounce ; -- nounce data
    wait4cipi                          ; -- wait for cipher to accept the nounce   
   -- Write key into the cipher
    cipi.mode    <= MODE_KEY           ; -- set mode to KEY accept
    cipi.valid   <= '1'                ; -- make the key valid for the cipher
    cipi.data    <= sim_key            ; -- key data
    wait4cipi                          ; -- wait for cipher to accept the key
   
   -- Write associated data blocks to the cipher
    for i in 0 to ADATA_BLOCKS-1 loop    
      cipi.mode    <= MODE_ADATA       ; -- set mode to Associate data
      cipi.valid   <= '1'              ; -- make the associate block valid for the cipher
      cipi.data    <= sim_adata(i)     ; -- associated data block
      cipi.len     <= sim_adata_len(i) ; -- length in bytes of the assoc. data block
      if i = ADATA_BLOCKS-1 then         -- raise the "last" control signal for the last block
        cipi.last  <= '1'              ;
      end if;
      wait4cipi                        ; -- wait for cipher to accept the associated data block                  
    end loop;
    cipi.last      <= '0'              ; -- "last" flag is deasserted after all the associated data blocks were exchanged
   
   -- Write plaintext data blocks to the cipher   
    for i in 0 to DATA_BLOCKS-1 loop    
      cipi.mode    <= MODE_DATA        ; -- set mode to data (plaintext)
      cipi.valid   <= '1'              ; -- make the plaintext block valid for the cipher      
      cipi.data    <= sim_ptext(i)     ; -- plaintext data block
      cipi.len     <= sim_data_len(i)  ; -- length in bytes of the plaintext block
      if i = DATA_BLOCKS-1 then          -- raise the "last" control signal for the last block
        cipi.last  <= '1'              ;
      end if;
      wait4cipi                        ; -- wait for cipher to accept the data block      
    end loop;  
    cipi.last      <= '0'              ; -- "last" flag is deasserted after all the plaintext blocks were exchanged

   -----------------------
   --- Decryption test ---
   -----------------------   
   cipi.decrypt  <= '1'                ; -- set decryption
   cipi.ready    <= '1'                ; -- TB can receive all the TAE output data immediately
   
   -- Write nonce into the cipher   
    cipi.mode    <= MODE_NOUNCE        ; -- set mode to NOUNCE
    cipi.valid   <= '1'                ; -- make the nounce valid for TAE
    cipi.data    <= sim_nounce         ; -- nounce data   
    wait4cipi                          ; -- wait for cipher to accept the nounce
   -- Write key into the cipher
    cipi.mode    <= MODE_KEY           ; -- set mode to KEY accept
    cipi.valid   <= '1'                ; -- make the key valid for the cipher
    cipi.data    <= sim_key            ; -- key data
    wait4cipi                          ; -- wait for cipher to accept the key
   
   -- Write associated data blocks to the cipher
    for i in 0 to ADATA_BLOCKS-1 loop    
      cipi.mode    <= MODE_ADATA       ; -- set mode to Associate data
      cipi.valid   <= '1'              ; -- make the associate block valid for the cipher
      cipi.data    <= sim_adata(i)     ; -- associated data block
      cipi.len     <= sim_adata_len(i) ; -- length in bytes of the assoc. data block
      if i = ADATA_BLOCKS-1 then         -- raise the "last" control signal for the last block
        cipi.last  <= '1'              ;
      end if;
      wait4cipi                        ; -- wait for cipher to accept the associated data block                  
    end loop;
    cipi.last      <= '0'              ; -- "last" flag is deasserted after all the associated data blocks were exchanged
   
   -- Write ciphertext data blocks to the cipher
    for i in 0 to DATA_BLOCKS-1 loop   
      cipi.mode    <= MODE_DATA        ; -- set mode to data (ciphertext)
      cipi.valid   <= '1'              ; -- make the ciphertext block valid for the cipher      
      cipi.data    <= sim_ctext(i)     ; -- ciphertext data block
      cipi.len     <= sim_data_len(i)  ; -- length in bytes of the ciphertext block
      if i = DATA_BLOCKS-1 then          -- raise the "last" control signal for the last block
        cipi.last  <= '1'              ;
      end if;
      wait4cipi                        ; -- wait for cipher to accept the data block      
    end loop;
    cipi.last      <= '0'              ; -- "last" flag is deasserted after all the ciphertext blocks were exchanged
    cipi.valid     <= '0'              ; -- no new data are valid for the cipher      
                                
   ---- END OF TEST ----
      
   wait;
end process;

-- Output monitoring and comparison
process(clk)
  variable monitor_rst: boolean := false;
  variable has_encdec : boolean := false;
  variable has_data   : natural range 0 to DATA_BLOCKS;
  variable cip_error  : boolean := false;
  variable decryption : boolean := false;  
begin   
  if rising_edge(clk) then        
      if cipi.reset = '1' or monitor_rst then                        
        monitor_rst := false;
        has_encdec  := false;
        has_data    := 0;
        cip_error   := false;        
        decryption  := false;        
      end if;
      
      -- Capture and compare tag 
      if has_encdec and (has_data = DATA_BLOCKS) and (cipo.valid and cipi.ready) = '1' then        
        report "Tag block output = 0x" & hstr(cipo.data); 
        if cipo.data = sim_tag then
          report "Tag is CORRECT!";
        else
          report "Tag is WRONG!";
          cip_error := true;
        end if;
        -- report if there were errors        
        if cip_error then
          if not decryption then
            report "!!!!ENCRYPTION FAILED!!!!";
          else
            report "!!!!DECRYPTION FAILED!!!!";
          end if;
        else
          if not decryption then
            report "!!!!ENCRYPTION SUCCESSFULLY COMPLETED!!!!";
          else
            report "!!!!DECRYPTION SUCCESSFULLY COMPLETED!!!!";
          end if;
        end if;
        -- Reset monitor
        monitor_rst := true;
      end if;
      
      -- Caputre and compare output associated data
      if has_encdec and (has_data < DATA_BLOCKS) and (cipo.valid and cipi.ready) = '1' then
        report "Data block output = 0x" & hstr(cipo.data); 
        if not decryption then
          -- encryption
          if cipo.data = sim_ctext(has_data) then
            report "Ciphertext block(" & integer'image(has_data) & ") is CORRECT!";
          else
            report "Ciphertext block(" & integer'image(has_data) & ") is WRONG!";
            cip_error := true;
          end if;          
        else
          -- decryption
          if cipo.data = sim_ptext(has_data) then
            report "Plaintext block(" & integer'image(has_data) & ") is CORRECT!";
          else
            report "Plaintext block(" & integer'image(has_data) & ") is WRONG!";
            cip_error := true;
          end if;          
        end if;
        
        -- Increment data block counter
        has_data := has_data + 1;
      end if;
      
      
      -- Capture encryption decryption
      if not has_encdec and (has_data = 0) and cipi.mode = MODE_ADATA and (cipi.valid and cipo.ready) = '1' then
        has_encdec := true;
        decryption := (cipi.decrypt = '1');
      end if;                 
  end if;
end process;

end;
