-------------------------------------------------------------------------------
--  128-bit SCREAM tweakable cipher implementation - Top level entity        --
-------------------------------------------------------------------------------
--  Author: L. Gaspar, TTTech Computertechnik AG, Vienna, Austria 2015(C)    --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This entity represents the top level SCREAM cipher module.  --
--               The core is designed to use logic resources only (suitable  --
--               for ASICs, too). The core features the 128-bit datapath that--
--               can compute either 1 round (folded) or 2 rounds (unrolled)  -- 
--               per clock cycle.                                            --
--                                                                           --
--  Data communication protocol: 
--               Data transfer handshaking is similar to ARM AXI standard.   --
--               When data/tweak/key are exchanged the sender indicates that --
--               that data are available on the bus by raising the valid     --
--               flag. Subsequently, receiver raises the ready flag to       --
--               acknowledge the data transfer. Thus, data are exchanged in  --
--               the cycle when both valid and ready flags are asserted.     --
--               In order to maximize performance, receiver can keep the     --
--               ready flag constantly asserted, so that data exchange occurs--
--               immediately after valid is raised. This way, no wait states --
--               occur and cipher can continue ciphering new data.           --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - PARAMETERS:                                                          --
--                                                                           --
--       - G_STEPS: number of cipher steps, where each step consists of two  --
--                  rounds. The recommended values are 10 (default) and 12.  --
--                                                                           --
--       - G_UNROLL: if True, 2 rounds (1 full step) are implemented.        --
--                   if False, only 1 round (1/2 of step) is implemented.    --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - DATA INPUT SIDE I/Os:                                                --
--                                                                           --
--       - decrypt_i: indicates if the exchanged input data are supposed to  --
--                    be encrypted (if LOW) or decrypted (if HIGH). This     -- 
--                    control bit is sampled during input data transfer      -- 
--       - data_i   : 128-bit key or plaintext(enc)/ciphertext(dec) input bus--
--       - tweak_i  : 128-bit tweak input bus                                --
--       - mode_i   : determines the type of data present on the data_i bus  --
--                    (0 = key data, 1 = plaintext data)                     --
--       - valid_i  : HIGH indicates that input data are valid on the data_i --
--                    bus and cipher can fetch them and start encryption     --
--       - ready_o  : is set HIGH by the cipher when input data were fetched --
--                                                                           --
--                                                                           --
--    - DATA OUTPUT SIDE I/Os:                                               --
--                                                                           --
--       - data_o  : 128-bit ciphertext(enc)/plaintext(dec) output bus       --
--       - valid_o : set to HIGH by the cipher when computation is finished  --
--                   and the resulting output data is present on the data_o  --
--       - ready_i : HIGH indicates that the data_o was successfully read    --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - Data organization:                                                   --
--        - Most significant bit:   [127]                                    --
--        - Most significant byte:  [127:120]                                --
--        - Least significant bit:  [0]                                      --
--        - Least significant byte: [7:0]                                    --
--                                                                           --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 12/12/2015 Lubos Gaspar: implementation of decryption             --
--       - 22/07/2014 Lubos Gaspar: first version of SCREAM                  --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.declaration_pkg.all;

-------------------------------------------------------------------------------
------------------------------- E N T I T Y -----------------------------------
-------------------------------------------------------------------------------
entity Scream128encdec is
  generic(
    G_STEPS    : natural := 10;    -- 10-12 steps (each step contains 2 rounds)
    G_UNROLL   : boolean := false  -- 1 (if false) or 2 (if true) rounds are impl.       
    );
  port (
    clk_i     : in  std_logic;    
    rst_i     : in  std_logic;
    -- Choose between encryption or decryption
    decrypt_i : in std_logic;
    -- Input interface side    
    valid_i   : in  std_logic;
    ready_o   : out std_logic;
    mode_i    : in  std_logic; -- 0 = key, 1 = data + tweak
    data_i    : in  std_logic_vector(127 downto 0); 
    tweak_i   : in  std_logic_vector(127 downto 0); 
    -- Output interface side
    valid_o   : out std_logic;
    ready_i   : in  std_logic;    
    data_o    : out std_logic_vector(127 downto 0)
    );
end Scream128encdec;

-------------------------------------------------------------------------------
------------------------- A R C H I T E C T U R E -----------------------------
-------------------------------------------------------------------------------
architecture rtl of Scream128encdec is

-- Each step consists of 2 rounds, thus cipher consists of STEP*2 rounds
constant ROUNDS      : natural := G_STEPS*2;

-- Round constant is defined in the SCREAM specifications v3
constant CONST       : natural := 2199;

-- Internal datapath signals
signal tweak_r      : std_logic_vector(127 downto 0);
signal tweak        : std_logic_vector(127 downto 0);
signal key_r        : std_logic_vector(127 downto 0);
signal key          : std_logic_vector(127 downto 0);
signal state_mux    : std_logic_vector(127 downto 0);
signal AddKey       : std_logic_vector(127 downto 0);
signal state        : std_logic_vector(127 downto 0);
signal state_tr     : std_logic_vector(127 downto 0);
signal sbox1_tr     : std_logic_vector(127 downto 0);
signal sbox1        : std_logic_vector(127 downto 0);
signal AddConstEnc1 : std_logic_vector(127 downto 0);
signal AddConstDec1 : std_logic_vector(127 downto 0);
signal llayer1      : std_logic_vector(127 downto 0);
signal illayer1     : std_logic_vector(127 downto 0);
signal illayer2     : std_logic_vector(127 downto 0);
signal sbox2_in_tr  : std_logic_vector(127 downto 0);
signal sbox2_tr     : std_logic_vector(127 downto 0);
signal sbox2        : std_logic_vector(127 downto 0);
signal AddConstEnc2 : std_logic_vector(127 downto 0);
signal AddConstDec2 : std_logic_vector(127 downto 0);
signal llayer2      : std_logic_vector(127 downto 0);
signal round_out    : std_logic_vector(127 downto 0);
signal cst          : natural range 0 to 2**16-1;

-- Datapath control signals
signal cnt          : natural range 0 to ROUNDS+1 := 0;
signal state_wr     : std_logic;
signal running      : std_logic;
signal done         : std_logic;
signal cnt_done     : std_logic;
signal start        : std_logic;
signal read_out     : std_logic;
signal decrypt      : std_logic;
signal decrypt_r    : std_logic;

---------------------------------------
-- Tweak computation functions
---------------------------------------
-- Calculate next tweak value
function twist_tweak (tweak: std_logic_vector(127 downto 0); 
                      mode: natural range 0 to 2)
return std_logic_vector is 
  variable tweak_low, tweak_high : std_logic_vector(63 downto 0);
  variable tweak_xor             : std_logic_vector(63 downto 0);
begin   
  -- Derive necessary half-words
  tweak_low  := tweak( 63 downto  0);
  tweak_high := tweak(127 downto 64);
  tweak_xor  := tweak_high xor tweak_low;  
  -- Calculate the next tweak value
  case ( mode ) is
    when 0 => return ( tweak_high & tweak_low  );
    when 1 => return ( tweak_xor  & tweak_high );
    when 2 => return ( tweak_low  & tweak_xor  );
  end case;  
end twist_tweak;

-- Select initial tweak
function init_tweak (tweak: std_logic_vector(127 downto 0); decrypt: std_logic)
return std_logic_vector is
begin
  if decrypt = '0' then
    -- Encryption uses the original tweak
    return twist_tweak ( tweak, 0 );
  else
    -- Decryption selects last encryption tweak
    return twist_tweak ( tweak, ROUNDS mod 3 );
  end if;
end init_tweak;

-- Tweak expansion function
function expand_tweak (tweak: std_logic_vector(127 downto 0); decrypt: std_logic)
return std_logic_vector is
begin
  if (G_UNROLL = true) then
    if decrypt = '0' then
      -- +2 tweak rounds % 3 = +2
      return twist_tweak ( tweak, 2 );
    else
      -- -2 tweak rounds % 3 = +1
      return twist_tweak ( tweak, 1 );
    end if;
  else
    if decrypt = '0' then
      -- +1 tweak round % 3 = +1
      return twist_tweak ( tweak, 1 );
    else
      -- -1 tweak round % 3 = +2
      return twist_tweak ( tweak, 2 );
    end if;
  end if;
end expand_tweak;

---------------------------------------
-- Initialization of the Round constant
---------------------------------------
function init_const (decrypt: std_logic) return natural is
begin     
  if decrypt = '0' then  
    -- Encryption uses the original constant
    return ( 0 );                                                            
  else
    -- Decryption uses the last encryption constant
    return ( ( (ROUNDS-1) * CONST ) mod 2**16 );
  end if;
end init_const;



-------------------------------------------------------------------------------
-------------------------- C I P H E R   B O D Y ------------------------------
-------------------------------------------------------------------------------
begin

-- Cipher start pulse 
start <= '1' when (valid_i = '1') and (mode_i = '1') and                         -- new data are available on input
                  (running = '0') and                                            -- cipher is not executing
                  (read_out = '1' or ((done = '1') and (ready_i = '1'))) else    -- result of the previous encryption was read out 
         '0';

-- Decryption/Encryption select
decrypt <= decrypt_i when (start = '1') else decrypt_r;

-- Cipher is ready to receive input data
ready_o <= '0' when running = '1' else                                           -- not ready if execution is in progress
           '0' when (done = '1') and (ready_i = '0') else                        -- not ready if previos result was not read out
           '1';

-- Output ciphertext is ready at the cipher data output bus
valid_o <= done;
           
---------------------
---- KEY STORAGE ----
---------------------
process(clk_i)
begin
  if rising_edge(clk_i) then
    if (valid_i = '1') and (mode_i = '0') and (running = '0') then               -- a new key is available on input and cipher is not executing
      key_r <= data_i;
    end if;
  end if;
end process;
key <= key_r;


---------------------
--- TWEAK STORAGE ---
---------------------
process(clk_i)
begin
  if rising_edge(clk_i) then
    -- Tweak expansion
    tweak_r  <=  expand_tweak(tweak,decrypt);    
  end if;
end process;

-- Use the original tweak in the first round
tweak <= init_tweak(tweak_i,decrypt) when running = '0' else tweak_r;

---------------------
---- MULTIPLEXER ----
---------------------
state_mux <= data_i when start = '1' else round_out;

---------------------
--- ADD ROUND KEY ---
---------------------
AddKey <= 
  -- When the step is unrolled take the "AddTweaKey" output
  key xor tweak xor state_mux when (G_UNROLL = true) else
  -- When the step is folded, take either "AddTweaKey" (in even rounds) 
  -- or previous state value (in odd rounds)
  key xor tweak xor state_mux when (cnt mod 2 = 0) or (running = '0') else  
  state_mux;

----------------------
--- STATE REGISTER ---
----------------------
-- Flip-flop state register
process(clk_i)
begin 
  if rising_edge(clk_i) then    
    if state_wr = '1' then
      state <= AddKey;
    end if;    
  end if;
end process;

---------- ROUND 1 ------------

----------------------------
--- INVERSE LINEAR LAYER ---
----------------------------
inv_lbox_layer1:
for i in 0 to 7 generate
  inv_lbox_inst1: lbox
    port map(
      decrypt_i => '1',
      data_i    =>    state((i+1)*16-1 downto i*16),
      data_o    => illayer1((i+1)*16-1 downto i*16)
    );
end generate inv_lbox_layer1;

-----------------------------------
---- ADD CONSTANT (DECRYPTION) ----
-----------------------------------
AddConstDec1(127 downto 16) <= illayer1(127 downto 16);
AddConstDec1( 15 downto  0) <= illayer1( 15 downto  0) xor std_logic_vector(to_unsigned(cst,16)) ;

---------------------------------
--- SBOX & INVERSE SBOX LAYER ---
---------------------------------
-- In case of encryption sbox operation is applied to the state. 
-- In case of decryption inverse sbox operation is applied after AddConstant.

-- 1) Transposition of the state matrix
state_tr <= Transpose(state,TR_DIRECT) when decrypt = '0' else
            Transpose(AddConstDec1,TR_DIRECT); 

-- 2) Perform Sbox transformation
sbox_layer1:
for i in 0 to 15 generate 
  sbox_inst1: sbox 
    port map(
      decrypt_i => decrypt,
      data_i    => state_tr((i+1)*8-1 downto i*8),
      data_o    => sbox1_tr((i+1)*8-1 downto i*8)
    );  
end generate sbox_layer1;

-- 3) Inverse transposition of the state matrix after applying Sbox transformation
sbox1 <= Transpose(sbox1_tr,TR_INVERSE);


-----------------------------------
---- ADD CONSTANT (ENCRYPTION) ----
-----------------------------------
AddConstEnc1(127 downto 16) <= sbox1(127 downto 16);
AddConstEnc1( 15 downto  0) <= 
   sbox1( 15 downto  0) xor std_logic_vector(to_unsigned(cst,16));

--------------------
--- LINEAR LAYER ---
--------------------
lbox_layer1:
for i in 0 to 7 generate
  lbox_inst1: lbox
    port map(
      decrypt_i => '0',
      data_i    => AddConstEnc1((i+1)*16-1 downto i*16),
      data_o    =>      llayer1((i+1)*16-1 downto i*16)
    );
end generate lbox_layer1;

-- Complete the cipher datapath loop for folded architecture
one_round:
if (G_UNROLL = false) generate
  round_out <= llayer1 when decrypt = '0' else sbox1;
end generate one_round;

---------- ROUND 2 ----------

two_rounds: 
if (G_UNROLL = true) generate

  ----------------------------
  --- INVERSE LINEAR LAYER ---
  ----------------------------
  inv_lbox_layer2:
  for i in 0 to 7 generate
    inv_lbox_inst2: lbox
      port map(
        decrypt_i => '1',
        data_i    =>    sbox1((i+1)*16-1 downto i*16),
        data_o    => illayer2((i+1)*16-1 downto i*16)
      );
  end generate inv_lbox_layer2;

  -----------------------------------
  ---- ADD CONSTANT (DECRYPTION) ----
  -----------------------------------
  AddConstDec2(127 downto 16) <= illayer2(127 downto 16);
  AddConstDec2( 15 downto  0) <= 
    illayer2( 15 downto  0) xor std_logic_vector(to_unsigned(cst-CONST,16));


  ---------------------------------
  --- SBOX & INVERSE SBOX LAYER ---
  ---------------------------------
  
  -- 1) Transposition of the state matrix
  sbox2_in_tr <= Transpose(llayer1,TR_DIRECT) when decrypt = '0' else
                 Transpose(AddConstDec2,TR_DIRECT); 
  
  -- 2) Perform Sbox transformation
  sbox_layer2:
  for i in 0 to 15 generate 
    sbox_inst2: sbox 
      port map(
        decrypt_i => decrypt,
        data_i    => sbox2_in_tr((i+1)*8-1 downto i*8),
        data_o    =>    sbox2_tr((i+1)*8-1 downto i*8)
      );  
  end generate sbox_layer2;  
  
  -- 3) Inverse transposition of the state matrix
  sbox2 <= Transpose(sbox2_tr,TR_INVERSE);
  
  -----------------------------------
  ---- ADD CONSTANT (ENCRYPTION) ----
  -----------------------------------
  AddConstEnc2(127 downto 16) <= sbox2(127 downto 16);  
  AddConstEnc2( 15 downto  0) <= 
    sbox2(15 downto 0) xor (std_logic_vector(to_unsigned(cst+CONST,16)));

  --------------------
  --- LINEAR LAYER ---
  --------------------  
  lbox_layer2:
  for i in 0 to 7 generate
    lbox_inst2: lbox
      port map(
        decrypt_i => '0',
        data_i    => AddConstEnc2((i+1)*16-1 downto i*16),
        data_o    =>   llayer2((i+1)*16-1 downto i*16)
      );    
  end generate lbox_layer2;
    
  -- Complete the cipher datapath loop for unrolled architecture
  round_out <= llayer2 when decrypt = '0' else sbox2;
  
end generate two_rounds;

----------------------
--- CIPHER OUTPUT  ---
----------------------
-- Content of the state register 
data_o <= state;
    
---------------------
--- CONTROL LOGIC ---
---------------------
  
-- Main state machine
process(clk_i, rst_i)
begin
  if rst_i = '1' then
    -- indicates if the cipher is encrypting or decrypting
    decrypt_r   <= '0';
    -- indicates if the cipher is running
    running     <= '0';
    -- indicates if the computation is finished
    done        <= '0';
    -- indicates if the output cipher text was read out
    read_out    <= '1';
  elsif rising_edge(clk_i) then
    -- Cipher starts
    if start = '1' then
      running   <= '1';
      done      <= '0';
      read_out  <= '0';
      decrypt_r <= decrypt_i;      
    -- Cipher finishes
    elsif cnt_done = '1' then
      running   <= '0';
      done      <= '1';
    end if;
    -- Cipher result exchanged
    if (done = '1') and (ready_i = '1') then
      read_out  <= '1';
      done      <= '0';
    end if;    
  end if;
end process;

-- State register write enable signal
state_wr <= running or start; 
  
-- Round counter
process(clk_i)
begin
  if rising_edge(clk_i) then            
    if (start = '0') then
      if running = '1' then                  
        if (G_UNROLL = false) then
          -- one round is performed per clock cycle
          cnt   <= cnt + 1;
        else      
          -- two rounds are performed per clock cycle
          cnt   <= cnt + 2;  
        end if;          
      end if;
    else
      -- Reset the counter at start
      cnt       <= 1;
    end if;
  end if;
end process;

-- Round counter done
cnt_done <= 
            '1' when not G_UNROLL and cnt = (ROUNDS-0) else -- 1-round(folded)
            '1' when     G_UNROLL and cnt = (ROUNDS-1) else -- 2-round(unrolled)
            '0';

-- Round constant accumulator
process(clk_i)
begin
  if rising_edge(clk_i) then
    if start = '1' then
      -- Constant counter reset
      cst <= init_const(decrypt);      
    else      
      if (G_UNROLL = true) then        
        if decrypt = '0' then
          -- Add the constant 2 times per step
          cst	<= (cst + 2*CONST) mod 2**16;
        else
          -- Substract the constant 2 times per step
          cst <= (cst - 2*CONST) mod 2**16;
        end if;
      else
        if decrypt = '0' then
          -- Add the constant 1 time per round
          cst	<= (cst + CONST) mod 2**16;
        else
          -- Subtract the constant 1 time per round
          cst	<= (cst - CONST) mod 2**16;
        end if;
      end if;
    end if;
  end if;
end process;

end rtl;