-------------------------------------------------------------------------------
--  128-bit SCREAM tweakable cipher implementation - TAE Top level entity    --
-------------------------------------------------------------------------------
--  Author: Stephanie Kerckhof, Universite Catholique de Louvain, 2015(C)    --
-------------------------------------------------------------------------------
--                                                                           --
--                                                                           --
-- This code implements the full mode version of Scream with a 128-bit data  --
-- path. The tweakable cipher used here implements a full round per clock    --
-- cycle.                                                                    --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
-- PARAMETERS:                                                               --
-- -----------                                                               --
--     TRUNC : allows to choose if the trucation of the last non full        --
--             output block will be done internally or externally.           --
--     PADD  : allows to choose if the padding is done internally or         --
--             externally.                                                   --
--     STEPS : number of cipher steps, where each step is composed of two    --
--             rounds. The recommended values are 10 (default) and 12.       --
--     UNROLL: if true, full step is implemented and can be executed in one  --    
--             clock cycle. Otherwise (if false), only 1 round (half of the  --
--             step) is implemented, thus encryption requires two times more --
--             clock cycles.                                                 --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
-- INPUT INTERFACE:                                                          --
-- ----------------                                                          --
--    - in_decrypt: indicates if the exchanged input data are supposed to be --
--                  encrypted (if LOW) or decrypted (if HIGH). This control  -- 
--                  bit is sampled when first associated data block is       --
--                  transferred to the cipher.                               -- 
--    - in_data   : 128-bit data input bus                                   -- 
--    - in_mode   : determines the type of data present on the in_data bus   --
--                  (0 = nounce, 1 = key, 2 = ass. data, 3 = ptext/ctext).   --
--    - in_length : gives the length of the input data in bytes, it is equal --
--                  to "10000" if the block is full.                         -- 
--    - in_last:   indicates when the last associated data or ptext/ctext    --
--                 are passed to the authenticated cipher.                   --
--    - in_valid:  in_data is valid when in_valid is high.                   --
--    - in_ready:  is set high by the authenticated cipher when in_data is   --
--                 fetched.                                                  --
--                                                                           --
-- Note: The user must send the data in the correct order (1. nounce, 2. key,--
--       3. ass.data, 4. ptexts/ctexts). If some types of data are missing,  -- 
--       the authenticated cipher will either take the previous provided     --
--       data (for key and nounce) or consider a null associated data.       --
--                                                                           --
--                                                                           --
-- OUTPUT INTERFACE:                                                         --
-- -----------------                                                         --
--    - out_data:   128-bit data output bus.                                 --
--    - out_mode:   determines the type of data present on the out_data      --
--                  bus (0 = ctext/ptext, 1 = tag).                          --
--    - out_length: gives the length of the output data in bytes, it is      --
--                  equal to "10000" if the block is full.                   --
--    - out_last:   indicates when the last ctext/ptext is output from the   --
--                  authenticated cipher.                                    --
--    - out_valid:  out_data is valid when out_valid is high.                --
--    - out_ready:  is set high by the user when out_data is fetched.        --
--                                                                           --
--                                                                           --
-- Note :                                                                    --
-- ------                                                                    --
-- in_data has the least significant byte at positions (127 downto 120).     --
-- E.g. : in_data <= 0x11223344556680000...00 for a padded associated        --
-- data. Similarly, the nounce is placed at the positions (127 downto 40);   --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--     - 13/12/2015 Lubos Gaspar: decryption implemented                     --
--     - 11/12/2015 Lubos Gaspar: I/O lengths entered in bytes (not bits)    --
--     - 18/08/2014 Stephanie Kerckhof: first version of the SCREAM TAE      --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <stephanie.kerckhof@uclouvain.be>, <lubos.gaspar@tttech.com>,     --
--         <fstandae@uclouvain.be>.                                          --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity Scream_tae is
   generic(
     TRUNC     : boolean := True;
     PADD      : boolean := True;
     STEPS     : natural := 12;
     --STEPS     : natural := 12;
     --UNROLL    : boolean := True
     UNROLL    : boolean := False
   );
   port (
     clk       : in  std_logic;
     rst       : in  std_logic;
     
     -- Input data side
     in_decrypt: in  std_logic; -- 0 = encryption, 1 = decryption
     in_valid  : in  std_logic;
     in_ready  : out std_logic;
     in_mode   : in  std_logic_vector(1 downto 0); -- 0 = nounce, 1 = key, 2 = ass data, 3 = ptext/ctext
     in_last   : in  std_logic;
     in_data   : in  std_logic_vector(127 downto 0);
     in_length : in  std_logic_vector(4 downto 0); -- in bytes

     -- Output data side
     out_valid : out std_logic;
     out_ready : in  std_logic;
     out_mode  : out std_logic; -- 0 = ctext/ptext, 1 = tag
     out_last  : out std_logic;
     out_length: out std_logic_vector(4 downto 0); -- in bytes
     out_data  : out std_logic_vector(127 downto 0) 
     );
end Scream_tae;


architecture rtl of Scream_tae is

function trunc_length2mask(mask_size:integer; data_length:std_logic_vector) 
return std_logic_vector;
function pad_length2mask(mask_size:integer; data_length:std_logic_vector) 
return std_logic_vector;

component Scream128encdec is
   generic(
     G_STEPS   : NATURAL := 10;
     G_UNROLL  : BOOLEAN := FALSE   
   );
   port (
     clk_i   : in  std_logic;     
     rst_i   : in  std_logic;
     -- Choose between encryption or decryption
     decrypt_i: in std_logic;
     -- Input interface side    
     valid_i : in  std_logic;
     ready_o : out std_logic;
     mode_i  : in  std_logic; -- 0 = key, 1 = data + tweak
     data_i  : in  std_logic_vector(127 downto 0); 
     tweak_i : in  std_logic_vector(127 downto 0);
     -- Output interface side
     valid_o : out std_logic;
     ready_i : in  std_logic;
     data_o  : out std_logic_vector(127 downto 0)
     );
end component;

-- Input block data/control 
signal ctrli_decrypt_r    : std_logic;
signal ctrli_block			  : std_logic_vector(127 downto 0);
signal ctrli_block_type		: std_logic;
signal ctrli_block_length	: std_logic_vector(6 downto 0);
signal ctrli_block_last		: std_logic;
signal ctrli_block_full		: std_logic;
signal ctrli_block_valid	: std_logic;
signal ctrli_block_ready	: std_logic;
signal ctrli_block_enbl		: std_logic;
-- Last input data flags
signal ctrli_in_data_last : std_logic;
signal ctrli_in_data_last_enbl: std_logic;
-- Input padding/trunking control
signal ctrli_pad_mask	  	: std_logic_vector(127 downto 0);
signal ctrli_trunc_mask		: std_logic_vector(127 downto 0);
-- Input nonce data/control
signal ctrli_nounce			  : std_logic_vector(87 downto 0);
signal ctrli_nounce_valid	: std_logic;
signal ctrli_nounce_ready	: std_logic;
signal ctrli_nounce_enbl	: std_logic;
-- Tag from the previous encryption/decryption is valid,
-- a new encryption/decryption can start
signal ctrli_tag_valid		: std_logic;
signal ctrli_tag_enbl		  : std_logic;
-- Tweak generation
signal ctw_nounce         : std_logic_vector(87 downto 0);
signal ctw_nounce_r			  : std_logic_vector(87 downto 0);
signal ctw_nounce_enbl		: std_logic;
signal ctw_mode				    : std_logic_vector(1 downto 0);
signal ctw_block_cnt		  : std_logic_vector(31 downto 0);
signal ctw_block_cnt_enbl	: std_logic;
signal ctw_block_cnt_rst	: std_logic;
signal ctw_tweak			    : std_logic_vector(127 downto 0);
signal ctw_tweak_enbl		  : std_logic;
-- Input ptext/ctext checksum
signal ccs_checksum			  : std_logic_vector(127 downto 0);
signal ccs_checksum_enbl	: std_logic;
signal ccs_checksum_rst		: std_logic;
signal ccs_in_data			  : std_logic_vector(127 downto 0);
-- Cihper core input data/control signals
signal lstc_in_decrypt    : std_logic;
signal lstc_in_data			  : std_logic_vector(127 downto 0);
signal lstc_in_data_r		  : std_logic_vector(127 downto 0);
signal lstc_in_data_be		: std_logic_vector(127 downto 0);
signal lstc_in_tweak		  : std_logic_vector(127 downto 0);
signal lstc_in_tweak_be		: std_logic_vector(127 downto 0);
signal lstc_in_mode			  : std_logic;
signal lstc_in_valid		  : std_logic;
signal lstc_in_ready		  : std_logic;
signal lstc_in_enbl			  : std_logic;
-- Cipher core output data/control signals
signal lstc_out_data_be		: std_logic_vector(127 downto 0);
signal lstc_out_data		  : std_logic_vector(127 downto 0);
signal lstc_out_valid		  : std_logic;
signal lstc_out_ready		  : std_logic;
signal lstc_out_enbl		  : std_logic;
signal lstc_out_reg_enbl	: std_logic;
signal lstc_out_data_r		: std_logic_vector(127 downto 0);
signal lstc_out_xor			  : std_logic_vector(127 downto 0);
signal lstc_out_auth		  : std_logic_vector(127 downto 0);
-- Truncated cipher output data
signal trunc_data			    : std_logic_vector(127 downto 0);

-- Output block data/control 
signal ctrlo_block_length	: std_logic_vector(6 downto 0);
signal ctrlo_block_type		: std_logic;
signal ctrlo_block_last		: std_logic;
signal ctrlo_block_full		: std_logic;
signal ctrlo_block_valid	: std_logic;
signal ctrlo_block_enbl		: std_logic;
signal ctrlo_block_busy		: std_logic;
-- Output tag flags
signal ctrlo_tag_busy		  : std_logic;
signal ctrlo_tag_enbl		  : std_logic;
-- Output associated data flags
signal ctrlo_adata_busy		: std_logic;
-- Output last ptext/ctext
signal ctrlo_in_data_last		: std_logic;

-- Function trunc_length2mask ------------------------------------------
-- Generates a mask that can be used to truncate a mask_size-bit
-- vector for which the useful bits are located in the left part.
-- The number of useful bits is given by the data_length vector.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--
function trunc_length2mask( mask_size : natural ; data_length: std_logic_vector)
return std_logic_vector is
	variable result : std_logic_vector(mask_size-1 downto 0);
begin 

  for i in 0 to (result'length-1) loop    
    if i < data_length then
      result(result'length-1-i) := '1';
    else
      result(result'length-1-i) := '0';
    end if;
  end loop;
	return result;
end function;
------------------------------------------------------------------------

-- Function pad_length2mask --------------------------------------------
-- Generates a mask that can be used to pad a mask_size-bit
-- vector for which the useful bits are located in the left part.
-- The number of useful bits is given by the data_length vector.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--
function pad_length2mask(mask_size : natural; data_length: std_logic_vector) 
return std_logic_vector is
	variable result : std_logic_vector(mask_size-1 downto 0);
begin
	result := (mask_size-1 downto 0 => '0');

	result(mask_size-1-conv_integer(unsigned(data_length))) := '1';

	return result;
end function;
------------------------------------------------------------------------

begin

-- Input interface (in) ------------------------------------------------
-- Generation of the input interface signals. in_ready is high when 
-- the LS Tweakable block Cipher (LSTC) is ready to receive a new data,
-- but not while the tag is generated. 
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--
in_ready	<= lstc_in_ready and not ctrli_tag_valid;

-- Storage of the decryption control signal
process(clk)
begin
  if rising_edge(clk) then       
    if ( -- if tae is running 
         ctrlo_block_busy or (ctrlo_tag_busy and
         -- if a new ptext/ctext was NOT received at the end of tag encryption
         not (in_valid and (lstc_in_ready and not ctrli_tag_valid)))) = '1' then       
      -- keep the value
      ctrli_decrypt_r <= ctrli_decrypt_r;
    else 
      -- otherwise (tae not running or a new ptext/ctext starts new operation)
      ctrli_decrypt_r <= in_decrypt;
    end if;
  end if;
end process;
------------------------------------------------------------------------

-- Control on input data (ctrli) ---------------------------------------
-- Generates the control signals that are needed to generate the inputs
-- of the LSTC. 
--    1. Controls related to the block of input data are generated. 
--       Block_type is 0 for associated data and 1 for ptext/ctext.
--    2. Padding of the last blocks of associated data or ptext/ctext.
--       Padding logic is only generated if PADD is true. 
--    3. Controls related to the nounce. The data present on in_data
--       is the nounce when in_mode = 0.
--    4. Controls that inform the presence of the last block of data
--       at input of the tweakable block cipher.
--    5. Controls informing that the tag is present at the input of
--       the tweakable block cipher.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--

-- Block of data ------------------

-- 0 = assoc. data; 1 = ptext/ctext
ctrli_block_type	 <= in_mode(0);
-- Convert block length from bytes to bits
ctrli_block_length <= in_length(3 downto 0) & "000"; 
-- Last block 
ctrli_block_last	 <= in_last;
-- Block is full
ctrli_block_full	 <= in_length(4) or not in_last;
-- input data are valid only if it is assoc.data / ptext / ctext
ctrli_block_valid	 <= in_valid and in_mode(1); 
-- block is ready to be processed by LSTC
ctrli_block_ready	 <= lstc_in_ready;
-- block starts to be processed
ctrli_block_enbl	 <= ctrli_block_valid and ctrli_block_ready; 

-- Padding -----------------------------

padd_gen : if PADD generate
	-- generate a truncation bit mask
  ctrli_trunc_mask <= trunc_length2mask(128, ctrli_block_length);
  -- generate a padding bit mask
	ctrli_pad_mask   <= pad_length2mask(128, ctrli_block_length);
	-- Process input data block
  ctrli_block	
        -- Truncate and padd the last incomplete (assoc. data block)
		<= (in_data and ctrli_trunc_mask) or ctrli_pad_mask 
				when (ctrli_block_last and not ctrli_block_full 
				      and not ctrli_block_type) = '1' else
			  -- Truncate the last incomplete (ptext / ctext data block)
       (in_data and ctrli_trunc_mask)
				when (ctrli_block_last and not ctrli_block_full
				      and ctrli_block_type) = '1' else
			  -- Keep the block if not last or if the block is complete
       (in_data);
end generate;

no_padd_gen : if not PADD generate
	-- If padding is disabled
  ctrli_block		<= in_data;
end generate;

-- Nounce ------------------------------
ctrli_nounce        <= in_data(127 downto 40);
-- a new nounce is valid 
ctrli_nounce_valid  <= in_valid and not in_mode(1) and not in_mode(0);
-- a new nounce is ready only when LSTC is idle
ctrli_nounce_ready  <= lstc_in_ready;
-- new nounce starts to be processed
ctrli_nounce_enbl   <= ctrli_nounce_valid and ctrli_nounce_ready;

-- Last ptext available at LSTC input --------
ctrli_in_data_last		<= ctrli_block_type and ctrli_block_last 
							and ctrli_block_valid;
ctrli_in_data_last_enbl	<= ctrli_in_data_last and lstc_in_ready;


-- Tag available at input of LSTC --------------
ctrli_tag_enbl	<= ctrli_tag_valid and lstc_in_ready;

process(clk)
begin
	if rising_edge(clk) then
		if (rst or ctrli_tag_enbl) = '1' then
			ctrli_tag_valid <= '0';
		elsif ctrli_in_data_last_enbl = '1' then
      ctrli_tag_valid <= '1';
		end if;
	end if;
end process;
------------------------------------------------------------------------

-- Compute tweak (ctw) -------------------------------------------------
-- Takes the nounce (N) from input interface and computes the tweak 
-- according to the specifications:
--   Ptext/Ctext -> Tc  = (N||block_cnt||0000 0000)
--   Ass. Data   -> Tc' = (N||block_cnt||0000 0100) when last block is complete
--                        (N||block_cnt||0000 0110) when last block is incomplete
--                        (N||block_cnt||0000 0010) else
--   Tag         -> TS  = (N||msg_lenght||1)
-- 
-- Tweak needs a block counter. This counter is updated each time
-- an associated data or a plaintext is enable at the input
-- (mode = 2 or 3). It is not updated on the last block of ptext / ctext
-- if it is incomplete.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--

-- Nounce ----------------
ctw_nounce       <= ctrli_nounce;
ctw_nounce_enbl  <= ctrli_nounce_enbl;

process(clk)
begin
	if rising_edge(clk) then
		if ctw_nounce_enbl = '1' then
			ctw_nounce_r	<= ctw_nounce;
		end if;
	end if;
end process;

-- Tweak mode ------------
-- (0 for associated data, 1 for ptext/ctext, 2 for tag)
ctw_mode(0)	<= ctrli_block_type;
ctw_mode(1) <= ctrli_tag_valid;

-- Tweak Block count------------

-- Count only complete (full) blocks
ctw_block_cnt_enbl  <=             
           ctrli_block_enbl and (not ctrli_block_last or ctrli_block_full);

-- Reset counter if:
ctw_block_cnt_rst   <= 
           -- last block of associated data is processed
          (ctrli_block_enbl and (ctrli_block_last and not ctw_mode(0)))
           -- or after tag processing starts
					 or (ctw_mode(1) and ctw_tweak_enbl)
           -- or if reset is asserted
					 or rst;

process(clk)
begin
	if rising_edge(clk) then
		if ctw_block_cnt_rst = '1' then
			ctw_block_cnt	<= x"00000000";
		else
			ctw_block_cnt	<= ctw_block_cnt + ctw_block_cnt_enbl;
		end if;
	end if;
end process;

-- Tweak------------------
-- tweak is composed of:
-- 1) nounce
ctw_tweak(127 downto 40)  <= ctw_nounce_r;
-- 2) block counter
ctw_tweak(39 downto 8)    <= ctw_block_cnt;
-- 3) control byte 
ctw_tweak(7 downto 0)     <=
		   -- In case of tag generation ( (last data block bit count || 1) << 1 )
       ctrlo_block_length & '1' when ctw_mode(1) = '1' else
       -- In case of last associated data block
		   "00000" & ctrli_block_last & (not ctrli_block_last or (ctrli_block_last and not ctrli_block_full)) &'0'
				when ctw_mode(0) = '0' else
       -- In case of all data blocks, and all (except last) associated data blocks
		   "00000000";

ctw_tweak_enbl	<= lstc_in_enbl;


------------------------------------------------------------------------

-- Compute Checksum (ccs) ----------------------------------------------
-- Generate the checksum from the input ptexts/ctexts.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	
ccs_checksum_rst	<= rst or ctrlo_tag_enbl;

-- Add checksum enable
ccs_checksum_enbl	<= 
                    -- Ptext is the LSTC input in case of encryption
                    (ctrli_block_enbl and ctrli_block_type) 
                     when ctrli_decrypt_r = '0' else
                    -- Ptext is the LSTC output in case of decryption 
                     (lstc_out_enbl and ctrlo_block_type);

-- Checksum accumulator input
ccs_in_data  <= 
                -- Ptext is the TAE input in case of encryption
                ctrli_block when ctrli_decrypt_r = '0' else                
                -- Ptext is the truncated LSTC output in case of decryption (last block only)
                trunc_data when (ctrlo_in_data_last and not ctrlo_block_full) = '1' else
                -- Ptext is the LSTC output in case of decryption
                lstc_out_data;

process(clk)
begin
	if rising_edge(clk) then
		if ccs_checksum_rst = '1' then
			ccs_checksum	<= (others => '0');
		elsif ccs_checksum_enbl = '1' then
			ccs_checksum	<= ccs_checksum xor ccs_in_data;
		end if;
	end if;
end process;
------------------------------------------------------------------------

-- LS Tweakable cipher (LSTC) ------------------------------------------
-- Instantiation of the LS tweakable block cipher. lstc_in interface
-- manages the inputs of the tweakable block cipher. in_data corresponds
-- to the checksum when the tag is generated, to the length of the
-- ptext/ctext if the last ptext/ctext to encrypt/decrypt is not a full block
-- or to the ptext/ctext data input by the user. The block cipher will consider
-- in_data either as a key or a ptext/ctext according to the mode 
-- (0 = key, 1 = ptext/ctext). 
--
-- lstc_out is used to output the corresponding ctext/ptext/tag. The outputs of
-- lstc will be remodeled according to the specifications and 
-- registered if needed. 
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--

-- LSTC input is valid
lstc_in_valid	<= (in_valid and not ctrli_nounce_valid) or ctrli_tag_valid;
-- LSTC accepted the data and starts encryption/decryption
lstc_in_enbl	<= lstc_in_valid and lstc_in_ready;
-- 0 = key on the LSTC input; 1 = data + tweak on LSTC input
lstc_in_mode	<= (in_mode(1) or ctrli_tag_valid);
-- LSTC input data
lstc_in_data	<= 
          -- checksum before being registered by CCS (in case of decryption only)
          ccs_checksum xor ccs_in_data when (ctrli_decrypt_r and ctrlo_in_data_last) = '1' else
          -- checksum already registered by CSS (in case of encryption only)
          ccs_checksum when ctrli_tag_valid = '1' else
					-- Length of the last ptext/ctext block if it is not complete
          (120 downto 0 => '0') & ctrli_block_length 
						when (ctrli_in_data_last and not ctrli_block_full) = '1' else
          -- TAE input block if the block is complete
					ctrli_block;

lstc_in_tweak	<= ctw_tweak;

-- LSTC mode:
     -- Decryption: 
     --   a) TAE decryption, only if the input ciphertext block is full (in case of TAE decryption)
     -- Encryption: if one of the following is true
     --   a) TAE encryption
     --   b) TAE decryption, when associated data or tag are processed
     --   c) TAE decryption, when last ciphertext block is incomplete
lstc_in_decrypt <= ctrli_decrypt_r and ctrli_block_type and ctrli_block_full and not ctrli_tag_enbl;

-- LSTC orders data in Big Endian format ([127:120] is the MSB, [7:0] is the LSB)
-- Since TAE uses Little Endian format ([127:120] is the LSB, [7:0] is the MSB)
-- data have to be remodeled to big endian before being processed by the LSTC
big_endian_gen : for i in 0 to 15 generate
	-- LSTC Input data or key
  lstc_in_data_be((i+1)*8-1 downto i*8)
           <= lstc_in_data(128-i*8-1 downto 128-(i+1)*8);
  -- LSTC Input tweak
	lstc_in_tweak_be((i+1)*8-1 downto i*8)
	         <= lstc_in_tweak(128-i*8-1 downto 128-(i+1)*8);
end generate; 

-- LSTC core
LScipher: Scream128encdec
   generic map(
     G_STEPS  => STEPS,
     G_UNROLL => UNROLL
   )
   port map(
     clk_i   => clk,     
     rst_i   => rst,
     
     decrypt_i => lstc_in_decrypt,
     -- Input interface side    
     valid_i => lstc_in_valid,
     ready_o => lstc_in_ready,
     mode_i  => lstc_in_mode,
     data_i  => lstc_in_data_be,
     tweak_i => lstc_in_tweak_be,

     -- Output interface side
     valid_o => lstc_out_valid,
     ready_i => lstc_out_ready,
     data_o  => lstc_out_data_be
     );     

-- LSTC output data is remodeled from Big Endian back to Little Endian format
little_endian_gen : for i in 0 to 15 generate
	lstc_out_data((i+1)*8-1 downto i*8)	
           <= lstc_out_data_be(128-i*8-1 downto 128-(i+1)*8);
end generate;

-- LSTC output ready
lstc_out_ready	<= out_ready or ctrlo_adata_busy;
-- LSTC output processed by the TAE
lstc_out_enbl	<= lstc_out_valid and lstc_out_ready;

-- Output mode computations ---------------

-- LSTC output is added to
lstc_out_xor	<= 
          -- TAE Input ptext/ctext (if last ptext/ctext block is incomplete)
          -- in case of ptext/ctext encryption/decryption
          lstc_out_data xor lstc_in_data_r
					when (ctrlo_in_data_last and not ctrlo_block_full) = '1' else
					-- Authentication data obtained from the associated data (Tag gener.)
          -- in case of tag generation
          lstc_out_data xor lstc_out_auth;

lstc_out_reg_enbl	<= lstc_out_enbl and ctrlo_adata_busy;

process(clk)
begin
	if rising_edge(clk) then
		-- TAE last input block storage
    if (rst or ctrlo_tag_enbl) = '1' then
      -- reset at the end of TAE operation (i.e. tag was generated)
			lstc_in_data_r	<= (others => '0');
		elsif ctrli_in_data_last = '1' then      
      -- store if input data block is incomplete 
      -- (needed later for being xored to LSTC output)
			lstc_in_data_r	<= in_data;
		end if;
		
    -- TAE Authentication data storage
		if (rst or ctrlo_tag_enbl) = '1' then
			-- reset at the end of TAE operation (i.e. tag was generated)
      lstc_out_auth	<= (others => '0');
		elsif lstc_out_reg_enbl = '1' then
      -- store when last associated data block was encrypted by LSTC
			lstc_out_auth	<= lstc_out_xor;
		end if;
	end if;
end process;
------------------------------------------------------------------------

-- Truncate last ctext/ptext (trunc) -----------------------------------
-- Will truncate the last ctext/ptext LSTC output block if it is not full.
-- The logic implementing this is only present if TRUNC is true.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--
trunc_gen_true : if TRUNC generate
	signal trunc_length	: std_logic_vector(7 downto 0);
	signal trunc_mask		: std_logic_vector(127 downto 0);
begin
	-- Number of bits to be truncated
  trunc_length	<= ctrlo_block_full & ctrlo_block_length;
	-- Truncation mask
  trunc_mask		<= trunc_length2mask(128,trunc_length);
  -- LSTC xored output after truncation
	trunc_data		<= lstc_out_xor and trunc_mask;
end generate;

trunc_gen_false : if not TRUNC generate
	-- if truncation is disabled
  trunc_data	<= lstc_out_xor;
end generate;
------------------------------------------------------------------------

-- Control on output data ----------------------------------------------
-- Generates the controls needed at the output of the LS tweakable
-- block cipher (LSTC). 
--    1. Controls related to the LSTC input block are registred in order
--       to keep the information for the output block.
--       Block_type is 0 for associated data and 1 for ptexts/ctexts.
--    2. Controls related to the tag computation are generated. 
--    3. Controls related to the associated data are generated.
--    4. Controls related to the ptext/ctext are generated.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--

-- Block control -------
process(clk)
begin
	if rising_edge(clk) then
		if ctrli_block_enbl = '1' then
			-- Store input control data
      ctrlo_block_type   <= ctrli_block_type;
			ctrlo_block_last   <= ctrli_block_last;
			ctrlo_block_full   <= ctrli_block_full;
			-- Zeros are added because input length is aligned to byte boundary
      ctrlo_block_length <= ctrli_block_length(6 downto 3) & "000";
		end if;
		
    -- Output busy flag
		if (rst or (ctrlo_block_busy and ctrlo_block_last and ctrlo_block_enbl and not ctrli_block_enbl)) = '1' then
      ctrlo_block_busy	<= '0';
		elsif ctrli_block_enbl = '1' then
			-- busy when TAE processes input associated data or input ptexts/ctexts
      -- not busy when during tag generation
      ctrlo_block_busy	<= '1';
		end if;
	end if;
end process;
-- LSTC output block is valid (corresponding to processed ptext/ctext or 
-- associated data, but not the encrypted checksum data for tag generation)
ctrlo_block_valid	<= lstc_out_valid and ctrlo_block_busy;
ctrlo_block_enbl	<= ctrlo_block_valid and lstc_out_ready;

-- Tag control ---------
process(clk)
begin
	if rising_edge(clk) then
		if (rst or ctrlo_tag_enbl) = '1' then 			
      ctrlo_tag_busy		<= '0';
		elsif ctrli_tag_enbl = '1' then
			-- High when tag is being generated
      ctrlo_tag_busy		<= '1';
		end if;
	end if;
end process;
ctrlo_tag_enbl	<= ctrlo_tag_busy and lstc_out_enbl;

-- Associated data control ------------
process(clk)
begin
	if rising_edge(clk) then
		if (rst or (ctrlo_block_enbl and ctrlo_block_last)) = '1' then
      ctrlo_adata_busy	<= '0';
		elsif (not ctrlo_adata_busy and not ctrli_block_type and ctrli_block_enbl)
      = '1' then
			-- High when associated data are processed
      ctrlo_adata_busy	<= '1';
		end if;
	end if;
end process;
						
-- Ptext/Ctext control -------
-- Last ctext/ptext data block processed by LSTC
ctrlo_in_data_last	<= ctrlo_block_type and ctrlo_block_last
						           and ctrlo_block_enbl;

------------------------------------------------------------------------

-- Output interface (out) ----------------------------------------------
-- Generates the data that are need for the output interface.
-- The data is valid when the data output from LSTC is a ctext/ptext 
-- that does not correspond to associated data or when the tag is generated.
-- Output data is either the result of a xor operation, the truncated data 
-- or the output of the LSTC.
--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--	--

-- TAE output data:
out_data	<= 
      -- LSTC output xored with authentication data (tag generation result)
      lstc_out_xor when ctrlo_tag_busy = '1' else
      -- Truncated XOR output (between LSTC output and input ptext/ctext)
      trunc_data when (ctrlo_in_data_last and not ctrlo_block_full) = '1' else
			-- Direct LSTC output (full ctext/ptext block only)
      lstc_out_data;

-- TAE output data valid flag
out_valid	<= lstc_out_valid and (ctrlo_block_type or ctrlo_tag_busy);
-- TAE output data mode (0 = ctext/ptext; 1 = tag)
out_mode	<= ctrlo_tag_busy;
-- TAE last output ctext/ptext block or tag
out_last	<= ctrlo_block_last;
-- Output length is in bytes, thus we have to truncate to (6 downto 3)
out_length	<= ctrlo_block_full & (ctrlo_block_length(6 downto 3));
------------------------------------------------------------------------

end rtl;
