-------------------------------------------------------------------------------
--  SCREAM declaration package                                               --
-------------------------------------------------------------------------------
--  Author: L. Gaspar, TTTech Computertechnik AG, Vienna, Austria 2015(C)    --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This package contains the necessary constants and functions --
--               to improve the readability of the VHDL code.                --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 13/12/2015 Lubos Gaspar: decryption support is added              --
--       - 22/07/2014 Lubos Gaspar: first version of SCREAM package          --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-------------------------------------------------------------------------------
-------------------- P A C K A G E   D E C L A R A T I O N --------------------
-------------------------------------------------------------------------------
package declaration_pkg is

-----------------------------
--- ENUMERATION CONSTANTS ---
-----------------------------  
constant TR_DIRECT  : natural := 0;
constant TR_INVERSE : natural := 1;

------------------------
--- TYPE DEFINITIONS ---
------------------------
type STATE_TYPE is array (0 to 7) of std_logic_vector(15 downto 0);

-----------------
--- FUNCTIONS ---
----------------- 

-- Transpose 16x8b state matrix into 8x16b or the other way around (if dir = 1)
function Transpose(x: std_logic_vector(127 downto 0); dir: natural)
return std_logic_vector;

-- Transforms 8x16-bit state matrix into 128-bit standard logic vector
function mx2vec(x: STATE_TYPE) return std_logic_vector;

-- Transforms 128-bit standard logic vector into 8x16-bit state matrix
function vec2mx(x: std_logic_vector(127 downto 0)) return STATE_TYPE;
  
------------------
--- COMPONENTS --- 
------------------

component sbox IS  
  port (
    decrypt_i : in  std_logic;
    data_i    : in  std_logic_vector(7 downto 0);
    data_o    : out std_logic_vector(7 downto 0)
    );
end component;

component lbox IS  
  port (
    decrypt_i : in  std_logic;
    data_i    : in  std_logic_vector(15 downto 0);
    data_o    : out std_logic_vector(15 downto 0)
    );
end component;

component Scream128encdec is
  generic(
    G_STEPS    : natural := 10;    -- 10-12 steps (each step contains 2 rounds)
    G_UNROLL   : boolean := false  -- 1 (if false) or 2 (if true) rounds are impl.    
    );
  port (
    clk_i     : in  std_logic;    
    rst_i     : in  std_logic;
    -- Choose between encryption or decryption
    decrypt_i : in  std_logic;
    -- Input interface side    
    valid_i   : in  std_logic;
    ready_o   : out std_logic;
    mode_i    : in  std_logic; -- 0 = key, 1 = data + tweak
    data_i    : in  std_logic_vector(127 downto 0); 
    tweak_i   : in  std_logic_vector(127 downto 0); 
    -- Output interface side
    valid_o   : out std_logic;
    ready_i   : in  std_logic;    
    data_o    : out std_logic_vector(127 downto 0)
    );
end component;

component Scream_tae is
   generic(
     TRUNC     : boolean := True;
     PADD      : boolean := True;
     STEPS     : natural := 20;
     UNROLL    : boolean := True
   );
   port (
     clk       : in  std_logic;
     rst       : in  std_logic;

     -- Input interface side
     in_decrypt: in  std_logic;
     in_valid  : in  std_logic;
     in_ready  : out std_logic;
     in_mode   : in  std_logic_vector(1 downto 0); -- 0 = nounce, 1 = key, 2 = ass data, 3 = ptext/ctext
     in_last   : in  std_logic;
     in_data   : in  std_logic_vector(127 downto 0); -- 127 = MSb
     in_length : in  std_logic_vector(4 downto 0); -- in bytes

     -- Output interface side
     out_valid : out std_logic;
     out_ready : in  std_logic;
     out_mode  : out std_logic; -- 0 = ctext/ptext, 1 = tag
     out_last  : out std_logic;
     out_length: out std_logic_vector(4 downto 0); -- in bytes
     out_data  : out std_logic_vector(127 downto 0)      
     );
end component;
  
end declaration_pkg;

-------------------------------------------------------------------------------
-------------------------- P A C K A G E   B O D Y ----------------------------
-------------------------------------------------------------------------------
package body declaration_pkg is  

-- Transpose 16x8b state matrix into 8x16b or the other way around (if dir = 1)
function Transpose(x: std_logic_vector(127 downto 0); dir: natural)
return std_logic_vector IS
  variable y : std_logic_vector(127 downto 0);
begin    
   if dir = TR_DIRECT then 
     -- transposing from 8x16b (state) to 16x8b (transposed state)
     for i in 0 to 7 loop
       for j in 0 to 15 loop
         y(j*8+i) := x(i*16+j);
       end loop;
     end loop;
   else            
     -- transposing back from 16x8b (transposed state) to 8x16b (state)
     for i in 0 to 15 loop
       for j in 0 to 7 loop
         y(j*16+i) := x(i*8+j);
       end loop;
     end loop;   
   end if;
 return y; 
end Transpose;

-- Transforms 8x16-bit state matrix into 128-bit standard logic vector
function mx2vec(x: STATE_TYPE)
return std_logic_vector is
  variable y : std_logic_vector(127 downto 0);  
begin           
  for i in 0 to 7 loop
    for j in 0 to 15 loop
      y(i*16+j) := x(i)(j);
    end loop;
  end loop;    
  return y;
end mx2vec;

-- Transforms 128-bit standard logic vector into 8x16-bit state matrix
function vec2mx(x: std_logic_vector(127 downto 0))
return STATE_TYPE is
  variable y : STATE_TYPE;   
begin           
  for i in 0 to 7 loop    
    for j in 0 to 15 loop           
      y(i)(j) := x(i*16+j);         
    end loop;    
  end loop;
  return y;
end vec2mx;

end declaration_pkg;
