-------------------------------------------------------------------------------
--  SCREAM 16-bit Lbox (linear layer)                                        --
-------------------------------------------------------------------------------
--  Author: L. Gaspar, TTTech Computertechnik AG, Wien, Austria, 2015        --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This entity represents the combinatory Lbox implementation  --
--               The core is generic - it does not utilize any embedded FPGA --
--               blocks, and thus it can also be synthesized for ASICs.      --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - I/Os:                                                                --
--       - decrypt_i : encryption(0) / decryption (1) selector               --
--       - data_i  : 16-bit Lbox input bus                                   --
--       - data_o  : 16-bit Lbox output bus                                  --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - Data organization:                                                   --
--        - Most significant bit:  [15]                                      --
--        - Least significant bit:  [0]                                      --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 05/12/2015 Lubos Gaspar: second version of the SCREAM Lbox        --
--       - 22/07/2014 Lubos Gaspar: first version of SCREAM Lbox             --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-------------------------------------------------------------------------------
------------------------------- E N T I T Y -----------------------------------
-------------------------------------------------------------------------------
entity lbox is
  port (
    decrypt_i : in  std_logic;
    data_i    : in  std_logic_vector(15 downto 0);
    data_o    : out std_logic_vector(15 downto 0)
    );
end lbox;

-------------------------------------------------------------------------------
------------------------- A R C H I T E C T U R E -----------------------------
-------------------------------------------------------------------------------
architecture rtl of lbox is

-- Declaration of the Lbox matrix type
type TYPE_LBOX is array (0 to 15) of std_logic_vector(0 to 15);

-- Definition of the Lbox matrix
constant Lbox_LUT : TYPE_LBOX := (
	"0011101000011100",
	"1001010101001010",
	"1100110111011110",
	"1000001101101001",
	"1011011011101011",
	"0000011101011100",
	"0010010010100111",
	"1010010101111111",
	"1101001001100010",
	"1111010110001111",
	"1100110010000101",
	"0010111011111110",
	"0100100011100110",
	"1111011001011110",
	"1101100000001110",
	"1000110101010001");

-- Definition of the Inverse Lbox matrix
constant ILbox_LUT : TYPE_LBOX := (
	"0111100111100111",
  "0010000011101110",
  "1000101101010100",
  "1100100011000110",
  "1010000000111011",
  "0110111101110101",
  "1001110010010100",
  "0111010101000001",
  "0010101001111000",
  "0111110110011101",
  "0001101110011000",
  "1010010100010101",
  "1111110101010110",
  "1010011101111110",
  "0110101111011110",
  "0001101101100001");
  
-------------------------------------------------------------------------------
-------------------------- C I P H E R   B O D Y ------------------------------
-------------------------------------------------------------------------------
begin

-- The Lbox is generated using the Lbox matrix (see SCREAM specs.)
-- The Lbox matrix specifies the positions of the XOR gates (see below)
process(data_i,decrypt_i)
  variable data : std_logic;
begin
  for j in 0 to 15 loop
    data := '0';
    for k in 0 to 15 loop
      if decrypt_i = '0' then
        -- Encryption
        data := data xor (data_i(k) and  Lbox_LUT(k)(j));
      else
        -- Decryption
        data := data xor (data_i(k) and ILbox_LUT(k)(j));
      end if;
    end loop;
    data_o(j) <= data;
  end loop;
end process;

end rtl;