-------------------------------------------------------------------------------
--  SCREAM 8-bit sbox (substitution layer)                                   --
-------------------------------------------------------------------------------
--  Author: L. Gaspar, TTTech Computertechnik AG, Wien, Austria, 2015        --
-------------------------------------------------------------------------------
--                                                                           --
--  Description: This entity represents the combinatory Sbox implementation  --
--               The core is generic - it does not utilize any embedded FPGA --
--               blocks, and thus it can also be synthesized for ASICs.      --
--               The core features 8-bit datapath and is purely combinatory. --
--               If G_BITSLICED is true, structure is more sequential with   --
--               longer critical path. If G_BITSLICED is false, structure is --
--               more parallel and has shorter critical path, but needs more --
--               resources.                                                  --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - I/Os:                                                                --
--       - decrypt_i : encryption(0) / decryption (1) selector               --
--       - data_i    : 8-bit Sbox input bus                                  --
--       - data_o    : 8-bit Sbox output bus                                 --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--    - Data organization:                                                   --
--        - Most significant bit:  [7]                                       --
--        - Least significant bit: [0]                                       --
--                                                                           --
-------------------------------------------------------------------------------
--  Revision history:                                                        --
--       - 04/12/2015 Lubos Gaspar: second version of the SCREAM Sbox        --
--       - 22/07/2014 Lubos Gaspar: first version of SCREAM Sbox             --
-------------------------------------------------------------------------------
--                                                                           --
-- Crypto group, UCL, Belgium, 2015.                                         --
-- e-mail: <lubos.gaspar@tttech.com>, <fstandae@uclouvain.be>.               --
--                                                                           --
-- To the extent possible under law, the author(s) have dedicated all        --
-- copyright and related and neighboring rights to this software to the      --
-- public domain worldwide. This software is distributed without any         --
-- warranty.                                                                 --
--                                                                           --
-- You should have received a copy of the CC0 Public Domain Dedication       --
-- along with this software. If not, see                                     --
-- <http://creativecommons.org/publicdomain/zero/1.0/>                       --
--                                                                           --
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


-------------------------------------------------------------------------------
------------------------------- E N T I T Y -----------------------------------
-------------------------------------------------------------------------------
entity sbox is
  generic (
    G_BITSLICED : boolean := true
    );
  port (
    decrypt_i : in  std_logic;
    data_i    : in  std_logic_vector(7 downto 0);
    data_o    : out std_logic_vector(7 downto 0)
    );
end sbox;

-------------------------------------------------------------------------------
------------------------- A R C H I T E C T U R E -----------------------------
-------------------------------------------------------------------------------
architecture rtl of sbox is

-- Define Sbox look-up table format
type TYPE_SBOX is array (0 to 255) of std_logic_vector(7 downto 0);

-- Direct Sbox look-up table
constant Sbox_LUT : TYPE_SBOX := 
    (
      X"20",X"8D",X"B2",X"DA",X"33",X"35",X"A6",X"FF",X"7A",X"52",X"6A",X"C6",X"A4",X"A8",X"51",X"23",
      X"A2",X"96",X"30",X"AB",X"C8",X"17",X"14",X"9E",X"E8",X"F3",X"F8",X"DD",X"85",X"E2",X"4B",X"D8",
      X"6C",X"01",X"0E",X"3D",X"B6",X"39",X"4A",X"83",X"6F",X"AA",X"86",X"6E",X"68",X"40",X"98",X"5F",
      X"37",X"13",X"05",X"87",X"04",X"82",X"31",X"89",X"24",X"38",X"9D",X"54",X"22",X"7B",X"63",X"BD",
      X"75",X"2C",X"47",X"E9",X"C2",X"60",X"43",X"AC",X"57",X"A1",X"1F",X"27",X"E7",X"AD",X"5C",X"D2",
      X"0F",X"77",X"FD",X"08",X"79",X"3A",X"49",X"5D",X"ED",X"90",X"65",X"7C",X"56",X"4F",X"2E",X"69",
      X"CD",X"44",X"3F",X"62",X"5B",X"88",X"6B",X"C4",X"5E",X"2D",X"67",X"0B",X"9F",X"21",X"29",X"2A",
      X"D6",X"7E",X"74",X"E0",X"41",X"73",X"50",X"76",X"55",X"97",X"3C",X"09",X"7D",X"5A",X"92",X"70",
      X"84",X"B9",X"26",X"34",X"1D",X"81",X"32",X"2B",X"36",X"64",X"AE",X"C0",X"00",X"EE",X"8F",X"A7",
      X"BE",X"58",X"DC",X"7F",X"EC",X"9B",X"78",X"10",X"CC",X"2F",X"94",X"F1",X"3B",X"9C",X"6D",X"16",
      X"48",X"B5",X"CA",X"11",X"FA",X"0D",X"8E",X"07",X"B1",X"0C",X"12",X"28",X"4C",X"46",X"F4",X"8B",
      X"A9",X"CF",X"BB",X"03",X"A0",X"FC",X"EF",X"25",X"80",X"F6",X"B3",X"BA",X"3E",X"F7",X"D5",X"91",
      X"C3",X"8A",X"C1",X"45",X"DE",X"66",X"F5",X"0A",X"C9",X"15",X"D9",X"A3",X"61",X"99",X"B0",X"E4",
      X"D1",X"FB",X"D3",X"4E",X"BF",X"D4",X"D7",X"71",X"CB",X"1E",X"DB",X"02",X"1A",X"93",X"EA",X"C5",
      X"EB",X"72",X"F9",X"1C",X"E5",X"CE",X"4D",X"F2",X"42",X"19",X"E1",X"DF",X"59",X"95",X"B7",X"8C",
      X"9A",X"F0",X"18",X"E6",X"C7",X"AF",X"BC",X"B8",X"E3",X"1B",X"D0",X"A5",X"53",X"B4",X"06",X"FE"
    );

-- Inverse Sbox look-up table
constant ISbox_LUT : TYPE_SBOX := 
    (
      X"8C",X"21",X"DB",X"B3",X"34",X"32",X"FE",X"A7",X"53",X"7B",X"C7",X"6B",X"A9",X"A5",X"22",X"50",
      X"97",X"A3",X"AA",X"31",X"16",X"C9",X"9F",X"15",X"F2",X"E9",X"DC",X"F9",X"E3",X"84",X"D9",X"4A",
      X"00",X"6D",X"3C",X"0F",X"38",X"B7",X"82",X"4B",X"AB",X"6E",X"6F",X"87",X"41",X"69",X"5E",X"99",
      X"12",X"36",X"86",X"04",X"83",X"05",X"88",X"30",X"39",X"25",X"55",X"9C",X"7A",X"23",X"BC",X"62",
      X"2D",X"74",X"E8",X"46",X"61",X"C3",X"AD",X"42",X"A0",X"56",X"26",X"1E",X"AC",X"E6",X"D3",X"5D",
      X"76",X"0E",X"09",X"FC",X"3B",X"78",X"5C",X"48",X"91",X"EC",X"7D",X"64",X"4E",X"57",X"68",X"2F",
      X"45",X"CC",X"63",X"3E",X"89",X"5A",X"C5",X"6A",X"2C",X"5F",X"0A",X"66",X"20",X"9E",X"2B",X"28",
      X"7F",X"D7",X"E1",X"75",X"72",X"40",X"77",X"51",X"96",X"54",X"08",X"3D",X"5B",X"7C",X"71",X"93",
      X"B8",X"85",X"35",X"27",X"80",X"1C",X"2A",X"33",X"65",X"37",X"C1",X"AF",X"EF",X"01",X"A6",X"8E",
      X"59",X"BF",X"7E",X"DD",X"9A",X"ED",X"11",X"79",X"2E",X"CD",X"F0",X"95",X"9D",X"3A",X"17",X"6C",
      X"B4",X"49",X"10",X"CB",X"0C",X"FB",X"06",X"8F",X"0D",X"B0",X"29",X"13",X"47",X"4D",X"8A",X"F5",
      X"CE",X"A8",X"02",X"BA",X"FD",X"A1",X"24",X"EE",X"F7",X"81",X"BB",X"B2",X"F6",X"3F",X"90",X"D4",
      X"8B",X"C2",X"44",X"C0",X"67",X"DF",X"0B",X"F4",X"14",X"C8",X"A2",X"D8",X"98",X"60",X"E5",X"B1",
      X"FA",X"D0",X"4F",X"D2",X"D5",X"BE",X"70",X"D6",X"1F",X"CA",X"03",X"DA",X"92",X"1B",X"C4",X"EB",
      X"73",X"EA",X"1D",X"F8",X"CF",X"E4",X"F3",X"4C",X"18",X"43",X"DE",X"E0",X"94",X"58",X"8D",X"B6",
      X"F1",X"9B",X"E7",X"19",X"AE",X"C6",X"B9",X"BD",X"1A",X"E2",X"A4",X"D1",X"B5",X"52",X"FF",X"07"
    );     

-------------------------------------------------------------------------------
-------------------------- C I P H E R   B O D Y ------------------------------
-------------------------------------------------------------------------------
begin

-- Bitsliced Sbox implementation results in more sequential stracture 
-- with longer critical path, but requires less FPGA/ASIC resources
Sbox_Bitsliced_impl: 
if G_BITSLICED generate
  -- Combinatorial Sbox   
  process(data_i, decrypt_i)
    variable w0,w1,w2,w3,w4,w5,w6,w7 : std_logic;
    variable t0,t1,t2,t3             : std_logic;
  begin 
    -- Assign all input bits
    w0               := data_i(0);
    w1               := data_i(1);
    w2               := data_i(2);
    w3               := data_i(3);
    w4               := data_i(4);
    w5               := data_i(5);
    w6               := data_i(6);
    w7               := data_i(7);
    
    -- Feistel 1st round  
    if decrypt_i = '0' then
      t0             :=     (w1 and w2) xor w0;
    else
      t0             := not((w1 and w2) xor w0);
    end if; 
    t1               := w1 xor w3;
    t2               := w2 xor t0;
    w4               := w4 xor ((w3 xor t2) and (w2 xor t1));
    w5               := w5 xor t2;
    w6               := w6 xor (w3 and t0);
    w7               := w7 xor (t1 and t2);
    
    -- Feistel 2nd round
    t0               := (w4 and w5) xor w6;
    t1               := (w5 or  w6) xor w7;
    t2               := (w7 and t0) xor w4;
    t3               := (w4 and t1) xor w5;
    w0               := w0 xor t0;
    w2               := w2 xor t1;
    w1               := w1 xor t2;
    w3               := w3 xor t3;
    
    -- Feistel 3rd round
    if decrypt_i = '0' then
      t0             := not((w1 and w2) xor w0);
    else
      t0             :=     (w1 and w2) xor w0;    
    end if; 
    t1               := w1 xor w3;
    t2               := w2 xor t0;
    w4               := w4 xor ((w3 xor t2) and (w2 xor t1));
    w5               := w5 xor t2;
    w6               := w6 xor (w3 and t0);
    w7               := w7 xor (t1 and t2);
    
    -- Assign all output bits
    data_o(0)        <= w0;
    data_o(1)        <= w1;
    data_o(2)        <= w2;
    data_o(3)        <= w3;
    data_o(4)        <= w4;
    data_o(5)        <= w5;
    data_o(6)        <= w6;
    data_o(7)        <= w7;
  end process;
end generate Sbox_Bitsliced_impl;


-- LUT Sbox implementation results in more parallel stracture 
-- with shorter critical path, but requires more FPGA/ASIC resources
Sbox_LUT_impl:
if not G_BITSLICED generate       
  data_o             <=  Sbox_LUT(to_integer(unsigned(data_i))) when decrypt_i = '0' else 
                        ISbox_LUT(to_integer(unsigned(data_i)));    
end generate Sbox_LUT_impl;  

end rtl;

